@@ -88,6 +88,18 @@ class MultiVector:
             self._algebra
         )
 
+    def __truediv__(self, other: self) -> self:
+        """Division, ie. multiplication with the inverse."""
+        assert isinstance(other, MultiVector)
+
+        return MultiVector(
+            self._algebra.geom_prod(
+                self._blade_values,
+                self._algebra.inverse(other._blade_values)
+            ),
+            self._algebra
+        )
+
     def __and__(self, other: self) -> self:
         """Regressive product. See `GeometricAlgebra.reg_prod()`"""
         assert isinstance(other, MultiVector)
@@ -104,7 +116,7 @@ class MultiVector:
             self._algebra
         )
 
-    def __negate__(self) -> self:
+    def __neg__(self) -> self:
         """Negation."""
         return MultiVector(
             -self._blade_values,
@@ -137,15 +149,15 @@ class MultiVector:
         )
 
     def __getitem__(self, key: Union[str, List[str]]) -> self:
-        """`MultiVector` with only passed blades as non-zeros."""
+        """`MultiVector` with only passed blade names as non-zeros."""
         return MultiVector(
-            self._algebra.keep_blades(self._blade_values, key),
+            self._algebra.keep_blades_with_name(self._blade_values, key),
             self._algebra
         )
 
     def __call__(self, key: Union[str, List[str]]):
-        """`tf.Tensor` with passed blades on last axis."""
-        return self._algebra.select_blades(self._blade_values, key)
+        """`tf.Tensor` with passed blade names on last axis."""
+        return self._algebra.select_blades_with_name(self._blade_values, key)
 
     def __repr__(self) -> str:
         return self._algebra.mv_repr(self._blade_values)
@@ -157,6 +169,13 @@ class MultiVector:
             self._algebra
         )
 
+    def simple_inverse(self) -> self:
+        """Simple inverse. See `GeometricAlgebra.simple_inverse()`."""
+        return MultiVector(
+            self._algebra.simple_inverse(self._blade_values),
+            self._algebra
+        )
+
     def dual(self) -> self:
         """Dual. See `GeometricAlgebra.dual()`."""
         return MultiVector(
@@ -178,20 +197,42 @@ class MultiVector:
             self._algebra
         )
 
-    def approx_exp(self) -> self:
+    def approx_exp(self, order: int = 50) -> self:
         """Approximate exponential. See `GeometricAlgebra.approx_exp()`."""
         return MultiVector(
-            self._algebra.approx_exp(self._blade_values),
+            self._algebra.approx_exp(self._blade_values, order=order),
             self._algebra
         )
 
-    def approx_log(self) -> self:
+    def exp(self, square_scalar_tolerance: Union[float, None] = 1e-4) -> self:
+        """Exponential. See `GeometricAlgebra.exp()`."""
+        return MultiVector(
+            self._algebra.exp(
+                self._blade_values,
+                square_scalar_tolerance=square_scalar_tolerance
+            ),
+            self._algebra
+        )
+
+    def approx_log(self, order: int = 50) -> self:
         """Approximate logarithm. See `GeometricAlgebra.approx_log()`."""
         return MultiVector(
-            self._algebra.approx_log(self._blade_values),
+            self._algebra.approx_log(self._blade_values, order=order),
             self._algebra
         )
 
     def is_pure_kind(self, kind: BladeKind) -> bool:
         """Whether the `MultiVector` is of a pure kind."""
         return self._algebra.is_pure_kind(self._blade_values, kind=kind)
+
+    def geom_conv1d(self, kernel: self,
+                    stride: int, padding: str,
+                    dilations: Union[int, None] = None) -> self:
+        """1D convolution. See `GeometricAlgebra.geom_conv1d().`"""
+        return MultiVector(
+            self._algebra.geom_conv1d(
+                self._blade_values, kernel._blade_values,
+                stride=stride, padding=padding, dilations=dilations
+            ),
+            self._algebra
+        )
