@@ -5,7 +5,7 @@ The `GeometricAlgebra` class is used to construct the algebra given a metric.
 It exposes methods for operating on `tf.Tensor` instances where their last
 axis is interpreted as blades of the algebra.
 """
-from typing import List, Any, Union
+from typing import List, Any, Union, Optional
 import numbers
 import tensorflow as tf
 import numpy as np
@@ -92,7 +92,7 @@ class GeometricAlgebra:
     @property
     def cayley(self) -> tf.Tensor:
         """`MxMxM` tensor where `M` is the number of basis
-        vectors in the algebra. Used for calculating the
+        blades in the algebra. Used for calculating the
         geometric product:
 
         `a_i, b_j, cayley_ijk -> c_k`
@@ -314,7 +314,10 @@ class GeometricAlgebra:
         blade_indices = tf.convert_to_tensor(blade_indices)
 
         # Don't allow duplicate indices
-        assert blade_indices.shape[0] == tf.unique(blade_indices)[0].shape[0]
+        tf.Assert(
+            blade_indices.shape[0] == tf.unique(blade_indices)[0].shape[0],
+            [blades]
+        )
 
         x = (
             tf.expand_dims(blade_signs, axis=-1) *
@@ -381,10 +384,10 @@ class GeometricAlgebra:
         tensor = tf.convert_to_tensor(tensor, dtype_hint=tf.float32)
         return self.grade_automorphism(self.reversion(tensor))
 
-    def inverse(self, a: tf.Tensor) -> tf.Tensor:
+    def simple_inverse(self, a: tf.Tensor) -> tf.Tensor:
         """Returns the inverted geometric algebra tensor
-        `X^-1` such that `X * X^-1 = 1` if
-        it exists.
+        `X^-1` such that `X * X^-1 = 1`. Only works for elements that
+        square to scalars. Faster than the general inverse.
 
         Args:
             a: Geometric algebra tensor to return inverse for
@@ -488,7 +491,7 @@ class GeometricAlgebra:
             a: Input geometric algebra tensor of shape
                 [..., Length, ChannelsIn, Blades]
             k: Geometric algebra tensor for the convolution kernel of shape
-                [ChannelsIn, ChannelsOut, KernelSize, Blades]
+                [KernelSize, ChannelsIn, ChannelsOut, Blades]
             stride: Stride to use for the convolution
             padding: "SAME" (zero-pad input length so output
                 length == input length / stride) or "VALID" (no padding)
@@ -544,6 +547,48 @@ class GeometricAlgebra:
             result += v / i_factorial
         return result
 
+    def exp(self, a: tf.Tensor, square_scalar_tolerance: Union[float, None] = 1e-4) -> tf.Tensor:
+        """Returns the exponential of the passed geometric algebra tensor.
+        Only works for multivectors that square to scalars.
+
+        Args:
+            a: Geometric algebra tensor to return exponential for
+            square_scalar_tolerance: Tolerance to use for the square scalar check
+                or None if the check should be skipped
+
+        Returns:
+            `exp(a)`
+        """
+        # See https://www.euclideanspace.com/maths/algebra/clifford/algebra/functions/exponent/index.htm
+        # for an explanation of how to exponentiate multivectors.
+
+        self_sq = self.geom_prod(a, a)
+
+        if square_scalar_tolerance is not None:
+            tf.Assert(tf.reduce_all(
+                tf.abs(self_sq[..., 1:]) < square_scalar_tolerance
+            ), [self_sq])
+
+        scalar_self_sq = self_sq[..., :1]
+
+        # "Complex" square root (argument can be negative)
+        s_sqrt = tf.sign(scalar_self_sq) * tf.sqrt(tf.abs(scalar_self_sq))
+
+        # Square to +1: cosh(sqrt(||a||)) + a / sqrt(||a||) sinh(sqrt(||a||))
+        # Square to -1: cos(sqrt(||a||)) + a / sqrt(||a||) sin(sqrt(||a||))
+        # TODO: Does this work for values other than 1 too? eg. square to +0.5?
+        # TODO: Find a solution that doesnt require calculating all possibilities
+        #       first.
+        non_zero_result = tf.where(
+            scalar_self_sq < 0,
+            (self.from_tensor(tf.cos(s_sqrt), [0]) +
+                a / s_sqrt * tf.sin(s_sqrt)),
+            (self.from_tensor(tf.cosh(s_sqrt), [0]) +
+                a / s_sqrt * tf.sinh(s_sqrt))
+        )
+
+        return tf.where(scalar_self_sq == 0, self.from_scalar(1.0) + a, non_zero_result)
+
     def approx_log(self, a: tf.Tensor, order: int = 50) -> tf.Tensor:
         """Returns an approximation of the natural logarithm using a centered
         taylor series. Only converges for multivectors where `||mv - 1|| < 1`.
@@ -595,7 +640,28 @@ class GeometricAlgebra:
             result = self.geom_prod(result, a)
         return result
 
-    def keep_blades(self, a: tf.Tensor, blade_names: Union[List[str], str]) -> tf.Tensor:
+    def keep_blades(self, a: tf.Tensor, blade_indices: List[int]) -> tf.Tensor:
+        """Takes a geometric algebra tensor and returns it with only the given
+        blade_indices as non-zeros.
+
+        Args:
+            a: Geometric algebra tensor to copy
+            blade_indices: Indices for blades to keep
+
+        Returns:
+            `a` with only `blade_indices` components as non-zeros
+        """
+        a = tf.convert_to_tensor(a, dtype_hint=tf.float32)
+        blade_indices = tf.cast(
+            tf.convert_to_tensor(blade_indices, dtype_hint=tf.int64),
+            dtype=tf.int64
+        )
+
+        blade_values = tf.gather(a, blade_indices, axis=-1)
+
+        return self.from_tensor(blade_values, blade_indices)
+
+    def keep_blades_with_name(self, a: tf.Tensor, blade_names: Union[List[str], str]) -> tf.Tensor:
         """Takes a geometric algebra tensor and returns it with only the given
         blades as non-zeros.
 
@@ -604,21 +670,39 @@ class GeometricAlgebra:
             blade_names: Blades to keep
 
         Returns:
-            `a` with only `blade_names` elements as non-zeros
+            `a` with only `blade_names` components as non-zeros
         """
-        a = tf.convert_to_tensor(a, dtype_hint=tf.float32)
-
         if isinstance(blade_names, str):
             blade_names = [blade_names]
 
         _, blade_indices = get_blade_indices_from_names(
             blade_names, self.blades)
 
-        blade_values = tf.gather(a, blade_indices, axis=-1)
+        return self.keep_blades(a, blade_indices)
 
-        return self.from_tensor(blade_values, blade_indices)
+    def select_blades(self, a: tf.Tensor, blade_indices: List[int]) -> tf.Tensor:
+        """Takes a geometric algebra tensor and returns a `tf.Tensor` with the
+        blades in blade_indices on the last axis.
 
-    def select_blades(self, a: tf.Tensor, blade_names: Union[List[str], str]) -> tf.Tensor:
+
+        Args:
+            a: Geometric algebra tensor to copy
+            blade_indices: Indices for blades to select
+
+        Returns:
+            `tf.Tensor` based on `a` with `blade_indices` on last axis.
+        """
+        a = tf.convert_to_tensor(a, dtype_hint=tf.float32)
+        blade_indices = tf.cast(
+            tf.convert_to_tensor(blade_indices, dtype_hint=tf.int64),
+            dtype=tf.int64
+        )
+
+        result = tf.gather(a, blade_indices, axis=-1)
+
+        return result
+
+    def select_blades_with_name(self, a: tf.Tensor, blade_names: Union[List[str], str]) -> tf.Tensor:
         """Takes a geometric algebra tensor and returns a `tf.Tensor` with the
         blades in blade_names on the last axis.
 
@@ -639,13 +723,43 @@ class GeometricAlgebra:
         blade_signs, blade_indices = get_blade_indices_from_names(
             blade_names, self.blades)
 
-        result = blade_signs * tf.gather(a, blade_indices, axis=-1)
+        result = blade_signs * self.select_blades(a, blade_indices)
 
         if is_single_blade:
             return result[..., 0]
 
         return result
 
+    def inverse(self, a: tf.Tensor) -> tf.Tensor:
+        """Returns the inverted geometric algebra tensor
+        `X^-1` such that `X * X^-1 = 1`.
+
+        Using Shirokov's inverse algorithm that works in arbitrary dimensions,
+        see https://arxiv.org/abs/2005.04015 Theorem 4.
+
+        Args:
+            a: Geometric algebra tensor to return inverse for
+
+        Returns:
+            inverted geometric algebra tensor
+        """
+        a = tf.convert_to_tensor(a, dtype_hint=tf.float32)
+
+        n = 2 ** ((len(self.metric) + 1) // 2)
+
+        u = a
+        for k in range(1, n):
+            c = n / k * self.keep_blades_with_name(u, "")
+            u_minus_c = u - c
+            u = self.geom_prod(a, u_minus_c)
+
+        if not self.is_pure_kind(u, BladeKind.SCALAR):
+            raise Exception(
+                "Can't invert multi-vector (det U not scalar: %s)." % u)
+
+        # adj / det
+        return u_minus_c / u[..., :1]
+
     def __call__(self, a: tf.Tensor) -> MultiVector:
         """Creates a `MultiVector` from a geometric algebra tensor.
         Mainly used as a wrapper for the algebra's functions for convenience.
